
# (C) Copyright 1996 University of Toronto
# Permission to use this software for non-commerical purposes is hereby
# granted.  This software is provided "as is" without express or implied
# warranty.  This software may not be transferred or sublicensed.
# Any enhancements made by the recipient of this software shall be
# shared with the original copyright holders.

# Acknowledgements
# This software was developed as part of the project entitled "Models
# for Supporting the Analysis and Redesign of Business Processes" funded
# by the Information Technology Research Centre of Ontario.
# Principal investigators for the project are John Mylopoulos, Ray
# Reiter, Hector Levesque, Eric Yu, and Yves Lesperance.
# The i* modelling framework (including the Strategic Dependency and
# Strategic Rationale models) was developed by Eric Yu in his doctoral
# dissertation, Department of Computer Science, University of Toronto,
# 1995.
# This software was originally written by Chen Wang as part of his
# Master degree requirement.
# Substantial modifications and enhancements were made by Godfrey Cheng
# and Seyil Yoon, who worked as summer students on this project.
# The "repository" part of the code was originally written by Brian Kramer and
# subsequently modified by Chen Wang, Seyil Yoon, and Godfrey Cheng.


#######
# Movement Routines
#######

proc do_move_from_link {canvas curtag l nx ny} {
    global Display_Node_Link PI  Display_Size

    if {[info exists Display_Node_Link($canvas,$curtag,$l)]} {
        foreach link $Display_Node_Link($canvas,$curtag,$l) {
        
            set coords [$canvas coords $link]
            set x1 [lindex $coords 0]
            set y1 [lindex $coords 1]
            set x2 [lindex $coords 2]
            set y2 [lindex $coords 3]

            $canvas coord $link $nx $ny \
                $x2 $y2

            if {$l == "Depend_From"} {
                set midx [expr ($nx+$x2)/2]
                set midy [expr ($ny+$y2)/2]

                set angle [compute_depend_stub_angle $nx $ny $x2 $y2]

                set cod [get_depend_stub_coord $midx $midy]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Depend_From) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3]

                $canvas itemconfigure \
                        $Display_Node_Link($canvas,$curtag,${link}_Depend_From) \
                        -start $angle

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Depend_Label_From)]} {
                    set cod2 [get_depend_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_Depend_Label_From) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }

            if {$l == "Task_From"} {
### en train de travailler by mike
    set distx [expr abs($x2-$x1)]
    set disty [expr abs($y2-$y1)]
    set dist [expr $distx+$disty]
    set width $Display_Size(Task)
#    puts $width
    if {$dist < 4*$width} {
	set ratio 0.55
	set dist2 [expr $ratio*$dist]
    } else {
	set dist2 [expr ($dist - $width)-20]
    }

    set alpha [expr $dist2/$dist]
    set beta [ expr 1- $alpha]
    set midx [expr $x1*$alpha+$x2*$beta]
    set midy [expr $y1*$alpha+$y2*$beta]



#    puts $width
#    if {$dist < 6*$width} {
#	set ratio 0.55
#    } elseif {$dist <  9*$width} {
#	set ratio 0.67
#    } elseif {$dist <  13*$width} {
#	set ratio 0.8
#    } else {
#	set ratio 0.9
#    }
#    set dist2 [expr $ratio*$dist]
#    set alpha [expr $dist2/$dist]
#    set beta [ expr 1- $alpha]
#    set midx [expr $x1*$alpha+$x2*$beta]
#    set midy [expr $y1*$alpha+$y2*$beta]
 
#    puts $x1,$x2,$midx
#    puts $y1,$y2,$midy
#   puts dist:,$dist,ratio:,$ratio
#                set midx [expr $nx*0.65+$x2*0.35]
#                set midy [expr $ny*0.65+$y2*0.35]

                set cod [get_task_stub_coord $midx $midy \
                         [compute_task_stub_slope $nx $ny $x2 $y2]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Task_From) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Task_Label_From)]} {
                    set cod2 [get_task_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_Task_Label_From) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }

            if {$l == "Means_From"} {
                set cod [get_means_stub_coord $nx $ny $x2 $y2 \
                         [compute_task_stub_slope $nx $ny $x2 $y2]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Means_From) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Means_Label_From)]} {
                    set cod2 [get_means_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_Means_Label_From) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }

            if {$l == "SGoal_From"} {
                set cod [get_softgoal_stub_coord $nx $ny $x2 $y2 \
                         [compute_task_stub_slope $nx $ny $x2 $y2]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_SGoal_From) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5] \
                        [lindex $cod 6] [lindex $cod 7]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_SGoal_Label_From)]} {
                    set cod2 [get_softgoal_label_coord $cod]

                    $canvas coord \
                      $Display_Node_Link($canvas,$curtag,${link}_SGoal_Label_From) \
                      [lindex $cod2 0] [lindex $cod2 1]
                }
            }
            
            if {$l == "Part_From"} {
                set cod [get_part_stub_coord $nx $ny $x2 $y2 \
                         [compute_task_stub_slope $nx $ny $x2 $y2]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Part_From) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Part_Label_From)]} {
                    set cod2 [get_part_label_coord $cod]

                    $canvas coord \
                      $Display_Node_Link($canvas,$curtag,${link}_Part_Label_From) \
                      [lindex $cod2 0] [lindex $cod2 1]
                }
            }
            
            if {$l == "Play_From"} {
                set cod [get_play_stub_coord $nx $ny $x2 $y2 \
                         [compute_task_stub_slope $nx $ny $x2 $y2]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Play_From) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Play_Label_From)]} {
                    set cod2 [get_play_label_coord $cod]

                    $canvas coord \
                      $Display_Node_Link($canvas,$curtag,${link}_Play_Label_From) \
                      [lindex $cod2 0] [lindex $cod2 1]
                }
            }
            
            if {$l == "Cover_From"} {
                set cod [get_cover_stub_coord $nx $ny $x2 $y2 \
                         [compute_task_stub_slope $nx $ny $x2 $y2]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Cover_From) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Cover_Label_From)]} {
                    set cod2 [get_cover_label_coord $cod]

                    $canvas coord \
                      $Display_Node_Link($canvas,$curtag,${link}_Cover_Label_From) \
                      [lindex $cod2 0] [lindex $cod2 1]
                }
            }
            
            if {$l == "Isa_From"} {
                set cod [get_isa_stub_coord $nx $ny $x2 $y2 \
                         [compute_task_stub_slope $nx $ny $x2 $y2]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Isa_From) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Isa_Label_From)]} {
                    set cod2 [get_isa_label_coord $cod]

                    $canvas coord \
                      $Display_Node_Link($canvas,$curtag,${link}_Isa_Label_From) \
                      [lindex $cod2 0] [lindex $cod2 1]
                }
            }

            if {$l == "Occupy_From"} {
                set cod [get_occupy_stub_coord $nx $ny $x2 $y2 \
                         [compute_task_stub_slope $nx $ny $x2 $y2]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Occupy_From) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Occupy_Label_From)]} {
                    set cod2 [get_occupy_label_coord $cod]

                    $canvas coord \
                      $Display_Node_Link($canvas,$curtag,${link}_Occupy_Label_From) \
                      [lindex $cod2 0] [lindex $cod2 1]
                }
            }
        }
    }
}


proc do_move_to_link {canvas curtag l nx ny} {
    global Display_Node_Link PI  Display_Size

    if {[info exists Display_Node_Link($canvas,$curtag,$l)]} {
        foreach link $Display_Node_Link($canvas,$curtag,$l) {
            set coords [$canvas coords $link]
            set x1 [lindex $coords 0]
            set y1 [lindex $coords 1]
            set x2 [lindex $coords 2]
            set y2 [lindex $coords 3]

            $canvas coord $link $x1 $y1 \
                $nx $ny

            if {$l == "Depend_To"} {
                set midx [expr ($x1+$nx)/2]
                set midy [expr ($y1+$ny)/2]

                set angle [compute_depend_stub_angle $x1 $y1 $nx $ny]

                set cod [get_depend_stub_coord $midx $midy]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Depend_To) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3]

                $canvas itemconfigure $Display_Node_Link($canvas,$curtag,${link}_Depend_To) \
                        -start $angle

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Depend_Label_To)]} {
                    set cod2 [get_depend_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_Depend_Label_To) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }

            if {$l == "Task_To"} {
### en train de travailler by mike

    set distx [expr abs($x2-$x1)]
    set disty [expr abs($y2-$y1)]
    set dist [expr $distx+$disty]
    set width $Display_Size(Task)
#    puts $width
    if {$dist < 4*$width} {
	set ratio 0.55
	set dist2 [expr $ratio*$dist]
    } else {
	set dist2 [expr ($dist - $width)-20]
    }

    set alpha [expr $dist2/$dist]
    set beta [ expr 1- $alpha]
    set midx [expr $x1*$alpha+$x2*$beta]
    set midy [expr $y1*$alpha+$y2*$beta]

#    set dist2 [expr $ratio*$dist]
#    set dist2 [expr ($dist - $width)-20]
#    set alpha [ expr $dist2/$dist]
#    set beta [expr 1- $alpha] 
#    puts $x1,$x2,$midx
#    puts $y1,$y2,$midy
#    puts dist:,$dist
#,ratio:,$ratio


#                set midx [expr $x1*0.65+$nx*0.35]
#                set midy [expr $y1*0.65+$ny*0.35]

                set cod [get_task_stub_coord $midx $midy \
                         [compute_task_stub_slope $x1 $y1 $nx $ny]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Task_To) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Task_Label_To)]} {
                    set cod2 [get_task_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_Task_Label_To) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }

            if {$l == "Means_To"} {
                set cod [get_means_stub_coord $x1 $y1 $nx $ny \
                         [compute_task_stub_slope $x1 $y1 $nx $ny]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Means_To) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Means_Label_To)]} {
                    set cod2 [get_means_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_Means_Label_To) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }

            if {$l == "SGoal_To"} {
                set cod [get_softgoal_stub_coord $x1 $y1 $nx $ny \
                         [compute_task_stub_slope $x1 $y1 $nx $ny]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_SGoal_To) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5] \
                        [lindex $cod 6] [lindex $cod 7]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_SGoal_Label_To)]} {
                    set cod2 [get_softgoal_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_SGoal_Label_To) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }
            
            if {$l == "Part_To"} {
                set cod [get_part_stub_coord $x1 $y1 $nx $ny \
                         [compute_task_stub_slope $x1 $y1 $nx $ny]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Part_To) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Part_Label_To)]} {
                    set cod2 [get_part_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_Part_Label_To) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }
            
            if {$l == "Play_To"} {
                set cod [get_play_stub_coord $x1 $y1 $nx $ny \
                         [compute_task_stub_slope $x1 $y1 $nx $ny]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Play_To) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Play_Label_To)]} {
                    set cod2 [get_play_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_Play_Label_To) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }

            if {$l == "Cover_To"} {
                set cod [get_cover_stub_coord $x1 $y1 $nx $ny \
                         [compute_task_stub_slope $x1 $y1 $nx $ny]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Cover_To) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Cover_Label_To)]} {
                    set cod2 [get_cover_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_Cover_Label_To) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }

            if {$l == "Isa_To"} {
                set cod [get_isa_stub_coord $x1 $y1 $nx $ny \
                         [compute_task_stub_slope $x1 $y1 $nx $ny]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Isa_To) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Isa_Label_To)]} {
                    set cod2 [get_isa_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_Isa_Label_To) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }

            if {$l == "Occupy_To"} {
                set cod [get_occupy_stub_coord $x1 $y1 $nx $ny \
                         [compute_task_stub_slope $x1 $y1 $nx $ny]]

                $canvas coord $Display_Node_Link($canvas,$curtag,${link}_Occupy_To) \
                        [lindex $cod 0] [lindex $cod 1] [lindex $cod 2] \
                        [lindex $cod 3] [lindex $cod 4] [lindex $cod 5]

                if {[info exists \
                        Display_Node_Link($canvas,$curtag,${link}_Occupy_Label_To)]} {
                    set cod2 [get_occupy_label_coord $cod]
                    $canvas coord \
                        $Display_Node_Link($canvas,$curtag,${link}_Occupy_Label_To) \
                        [lindex $cod2 0] [lindex $cod2 1]
                }
            }
        }
    }
}


############################################################
# Selection Routines
############################################################

proc select_second_node {canvas {done ""} {sign ""}} {
    global first_tag Display_Node_Tag
    
    set id [$canvas find withtag current]
    if {[info exists Display_Node_Tag($canvas,$id)]} {
        set second_tag $Display_Node_Tag($canvas,$id)

        $canvas bind node <Button-1> "no_op"

        if {$done != ""} {
            $done $canvas $first_tag($canvas) $second_tag $sign
        }

        set first_tag($canvas) ""
    }
}

proc select_first_node {canvas {done ""} {before_second ""} {sign ""}} {
    global first_tag Display_Node_Tag

    set id [$canvas find withtag current]

    if {[info exists Display_Node_Tag($canvas,$id)]} {
        set first_tag($canvas) $Display_Node_Tag($canvas,$id)

        if {$before_second != ""} {
            $before_second $canvas
        }

        $canvas bind node <Button-1> "
            select_second_node $canvas $done $sign
        "
    }
}

proc additional_link {canvas {done ""} {before_first ""} \
    {before_second ""} {sign ""}} {
    
    if {$before_first != ""} {
        $before_first $canvas
    }
    
    $canvas bind node <Button-1> \
        "select_first_node $canvas $done $before_second $sign"
}

############################################################################
# 
############################################################################

proc post_link_menu {canvas m x y} {
    catch {
        focus $m
        $m post [expr $x + [winfo rootx $canvas]] \
               [expr $y + [winfo rooty $canvas]]
        grab $canvas
    } msg
}

proc unpost_link_menu {canvas m} {
    catch {
        $m unpost
        grab release $canvas
        focus $canvas
    }
}

proc create_link_menu {canvas link} {
   set menu [menu $canvas.${link}_menu]

   $menu add command -label "Dismiss" -command \
           "unpost_link_menu $canvas $menu"
    
   $menu add separator

   $menu add command -label "Delete Current Link" -command \
           "menu_delete_current_link $canvas $menu $link"
   
   $menu add separator

   return $menu
}

############################################################################
# Deleting Links
############################################################################

proc delete_current_link_from {canvas link type tag} {
    global Display_Node_Link

    if {[info exists Display_Node_Link($canvas,$tag,${type}_From)]} {
        if {[llength $Display_Node_Link($canvas,$tag,${type}_From)] == 1} {
            unset Display_Node_Link($canvas,$tag,${type}_From)
        } else {
            set i [lsearch $Display_Node_Link($canvas,$tag,${type}_From) $link]
            set Display_Node_Link($canvas,$tag,${type}_From) \
                    [lreplace $Display_Node_Link($canvas,$tag,${type}_From) $i $i]
        }
    }

    if {[info exists Display_Node_Link($canvas,$tag,${link}_${type}_From)]} {
        set stub $Display_Node_Link($canvas,$tag,${link}_${type}_From)
        $canvas delete $stub
        unset Display_Node_Link($canvas,$tag,${link}_${type}_From)
    }

    if {[info exists Display_Node_Link($canvas,$tag,${link}_${type}_Label_From)]} {
        set l $Display_Node_Link($canvas,$tag,${link}_${type}_Label_From)
        $canvas delete $l
        unset Display_Node_Link($canvas,$tag,${link}_${type}_Label_From)
    }
}

proc delete_current_link_to {canvas link type tag} {
    global Display_Node_Link

    if {[info exists Display_Node_Link($canvas,$tag,${type}_To)]} {
        if {[llength $Display_Node_Link($canvas,$tag,${type}_To)] == 1} {
            unset Display_Node_Link($canvas,$tag,${type}_To)
        } else {
            set i [lsearch $Display_Node_Link($canvas,$tag,${type}_To) $link]
            set Display_Node_Link($canvas,$tag,${type}_To) \
                    [lreplace $Display_Node_Link($canvas,$tag,${type}_To) $i $i]
        }
    }

    if {[info exists Display_Node_Link($canvas,$tag,${link}_${type}_To)]} {
        set stub $Display_Node_Link($canvas,$tag,${link}_${type}_To)
        $canvas delete $stub
        unset Display_Node_Link($canvas,$tag,${link}_${type}_To)
    }

    if {[info exists Display_Node_Link($canvas,$tag,${link}_${type}_Label_To)]} {
        set l $Display_Node_Link($canvas,$tag,${link}_${type}_Label_To)
        $canvas delete $l
        unset Display_Node_Link($canvas,$tag,${link}_${type}_Label_To)
    }
}

proc delete_display_link {canvas link} { 
    global Display_Links Display_Link_Type

    if {$link != ""} {
        $canvas delete $link

        if {[info exists Display_Links($canvas,$link)]} {
            set tag1 [lindex $Display_Links($canvas,$link) 0]
            set tag2 [lindex $Display_Links($canvas,$link) 1]
            set type $Display_Link_Type($canvas,$link)
        
            delete_current_link_from $canvas $link $type $tag1
            delete_current_link_to $canvas $link $type $tag2
        }
    }
}

proc delete_current_link {canvas} {
    set link [$canvas find withtag current]

    delete_display_link $canvas $link
}

proc menu_delete_current_link {canvas menu link} {
    unpost_link_menu $canvas $menu

    delete_display_link $canvas $link
}

############################################################################
# Drawing Routines
############################################################################

proc bind_line {canvas link} {
    set menu [create_link_menu $canvas $link]

    lower_canvas_background $canvas

    $canvas bind $link <ButtonPress-3> "
        post_link_menu $canvas $menu %x %y
    "

    $canvas bind $link <ButtonRelease-3> "
        unpost_link_menu $canvas $menu
    "
}

proc add_to_node_link {canvas tag l link} {
    global Display_Node_Link

    if {[info exists Display_Node_Link($canvas,$tag,$l)] } {
        lappend Display_Node_Link($canvas,$tag,$l) $link
    } else {
        set Display_Node_Link($canvas,$tag,$l) "$link"
    }
}

proc add_to_link {canvas link type tag1 tag2} {
    global Display_Links Display_Link_Type

    set Display_Links($canvas,$link) "$tag1 $tag2"
    set Display_Link_Type($canvas,$link) $type
}

proc get_link_tags {tag1 tag2} {
   return [list \
           link$tag1$tag2 from$tag1 to$tag2 link item]
}

proc get_internal_link_tags {tag1 tag2} {
   set t [get_link_tags $tag1 $tag2]
   lappend t $tag1 $tag2
   return $t
}

proc get_stub_tags {link} {
   return [list \
           stub$link stub item]
}

proc get_link_label_tags {link} {
   return [list \
           linklabel$link linklabel item]
}

########################################################################
# Start Links
########################################################################

proc get_depend_link_tags {tag1 tag2} {
    set t [get_link_tags $tag1 $tag2]
    lappend t dependlink$tag1$tag2 dependlink
    return $t
}

proc get_depend_stub_tags {link} {
    set t [get_stub_tags $link]
    lappend t dependstub$link dependstub
    return $t
}

proc get_depend_link_label_tags {link} {
    set t [get_link_label_tags $link]
    lappend t dependlinklabel$link dependlinklabel
    return $t
}

proc get_depend_stub_coord {x y} {
    global Display_Link_Size
    
    set stub $Display_Link_Size(Depend)

    return [list [expr $x-$stub] [expr $y-$stub] \
            [expr $x+$stub] [expr $y+$stub]]
}

proc compute_depend_stub_angle {x1 y1 x2 y2} {
    global PI
    if {[expr $x2-$x1] != 0} {
        set angle [expr 90-atan(($y2*1.0-$y1*1.0)/($x2*1.0-$x1*1.0))/$PI*180]
        if {[expr $x2 > $x1] != 0} {
            set angle [expr $angle+180]
        }
    } else {
        set angle 0.0
    }

    return $angle
}

proc get_depend_label_coord {coord} {
    return [list [lindex $coord 0] [lindex $coord 1]]
}

proc create_depend_link {canvas tag1 tag2 {label ""} {menu 1}} {
    global Display_Node_Pos Display_Node_Link PI Display_Generic
    global Display_Font

    set fn $Display_Font(Link)

    set x1 [lindex $Display_Node_Pos($canvas,$tag1) 0]
    set y1 [lindex $Display_Node_Pos($canvas,$tag1) 1]

    set x2 [lindex $Display_Node_Pos($canvas,$tag2) 0]
    set y2 [lindex $Display_Node_Pos($canvas,$tag2) 1]
    
    set color $Display_Generic(color)

    set midx [expr ($x1+$x2)/2]
    set midy [expr ($y1+$y2)/2]

    set angle [compute_depend_stub_angle $x1 $y1 $x2 $y2]

    set link \
      [$canvas create line $x1 $y1 \
        $x2 $y2 -width 2 \
        -fill black \
        -tag [get_depend_link_tags $tag1 $tag2]]

    set coords [get_depend_stub_coord $midx $midy]

    set tmp \
      [$canvas create arc [lindex $coords 0] [lindex $coords 1] \
        [lindex $coords 2] [lindex $coords 3] \
        -extent 180 -start $angle -width 2 -fill $color -outline black \
        -tag [get_depend_stub_tags $link]]

    if {$label != ""} {
        set cod [get_depend_label_coord $coords]
        set tmp2 \
                [$canvas create text [lindex $cod 0] [lindex $cod 1] \
                -text "$label" -anchor s -font $fn\
                -tag [get_depend_link_label_tags $link]]

        $canvas lower $tmp2

        add_to_node_link $canvas $tag1 ${link}_Depend_Label_From $tmp2
        add_to_node_link $canvas $tag2 ${link}_Depend_Label_To $tmp2
    }

    $canvas lower $tmp
    $canvas lower $link

    add_to_node_link $canvas $tag1 Depend_From $link
    add_to_node_link $canvas $tag2 Depend_To $link
    add_to_node_link $canvas $tag1 ${link}_Depend_From $tmp
    add_to_node_link $canvas $tag2 ${link}_Depend_To $tmp
    
    add_to_link $canvas $link Depend $tag1 $tag2

    if {$menu == 1} {
        bind_line $canvas $link
    } else {
        lower_canvas_background $canvas
    }
}

proc get_task_link_tags {tag1 tag2} {
    set t [get_link_tags $tag1 $tag2]
    lappend t dependlinklabel$tag1$tag2 dependlinklabel
    return $t
}

proc get_task_stub_tags {link} {
    set t [get_stub_tags $link] 
    lappend t taskstub$link taskstub
    return $t
}

proc get_task_link_label_tags {link} {
    set t [get_link_label_tags $link] 
    lappend t tasklinklabel$link tasklinklabel
    return $t
}

proc compute_task_stub_slope {x1 y1 x2 y2} {
    return [expr ($y2*1.0-$y1*1.0)/($x2*1.0-$x1*1.0+0.001)]
}

proc get_task_stub_coord {x y slope} {
    global Display_Link_Size PI
    
    set tmp $Display_Link_Size(Task)

    if {[expr $slope-0.0] != 0} {
        set dx [expr sqrt($tmp*$tmp/(1/(($slope*1.0)*$slope*1.0+0.0001)+1+0.001))]
        set dy [expr (0-1/($slope*1.0+0.001))*$dx]
    } else {
        set dx 0
        set dy $tmp
    }

    return [list [expr $x-$dx] [expr $y-$dy] [expr $x+$dx] [expr $y+$dy]]
}

proc get_task_label_coord {coord} {
    return [list [expr \
            [expr [lindex $coord 0] < [lindex $coord 2]] ? \
             [lindex $coord 0] : [lindex $coord 2]] \
           [expr \
            [expr [lindex $coord 1] < [lindex $coord 3]] ? \
             [lindex $coord 1] : [lindex $coord 3]]]
}

proc create_task_decomp_link {canvas tag1 tag2 {label ""}} {
    global Display_Node_Pos Display_Node_Link
    global Display_Font Display_Size

    set fn $Display_Font(Link)

    set x1 [lindex $Display_Node_Pos($canvas,$tag1) 0]
    set y1 [lindex $Display_Node_Pos($canvas,$tag1) 1]

    set x2 [lindex $Display_Node_Pos($canvas,$tag2) 0]
    set y2 [lindex $Display_Node_Pos($canvas,$tag2) 1]

    set link \
      [$canvas create line $x1 $y1 \
        $x2 $y2 -width 2 \
        -fill black -tag [get_task_link_tags $tag1 $tag2]]

### changed  by mike
    set distx [expr abs($x2-$x1)]
    set disty [expr abs($y2-$y1)]
    set dist [expr $distx+$disty]
 
   set width $Display_Size(Task)
#    puts $width
    if {$dist < 6*$width} {
	set ratio 0.55
	set dist2 [expr $ratio*$dist]
   } else {
	set dist2 [expr ($dist - $width)-20.0]
#	puts $dist,$dist2
    }

    set alpha [expr $dist2/$dist]
    set beta [ expr 1- $alpha]
    set lx [expr $x1*$alpha+$x2*$beta]
    set ly [expr $y1*$alpha+$y2*$beta]
#    puts alpha:$alpha,beta:$beta
#    puts X1:$x1,Y1:$y1
#    puts X2:$x2,Y2:$y2
#    puts lx:$lx,ly:$ly

#    set width $Display_Size(Task)
#    puts $width
 #   if {$dist < 6*$width} {
#	set ratio 0.55
#    } elseif {$dist <  9*$width} {
#	set ratio 0.67
#    } elseif {$dist <  13*$width} {
#	set ratio 0.8
#    } else {
#	set ratio 0.9
#    }
#    set dist2 [expr $ratio*$dist]
#    set alpha [expr $dist2/$dist]
#    set beta [ expr 1- $alpha]
#    set lx [expr $x1*$alpha+$x2*$beta]
#    set ly [expr $y1*$alpha+$y2*$beta]
#    set lx [expr $x1*0.65+$x2*0.35]
#    set ly [expr $y1*0.65+$y2*0.35]

    set coord [get_task_stub_coord $lx $ly \
            [compute_task_stub_slope $x1 $y1 $x2 $y2]]

    set tmp \
      [$canvas create line [lindex $coord 0] \
       [lindex $coord 1] [lindex $coord 2] [lindex $coord 3] -width 2 \
       -fill black -tag [get_task_stub_tags $link]]

    if {$label != ""} {
        set cod [get_task_label_coord $coord]

        set tmp2 \
                [$canvas create text [lindex $cod 0] [lindex $cod 1] \
                -text "$label" -anchor s -font $fn\
                -tag [get_task_link_label_tags $link]]

        $canvas lower $tmp2

        add_to_node_link $canvas $tag1 ${link}_Task_Label_From $tmp2
        add_to_node_link $canvas $tag2 ${link}_Task_Label_To $tmp2
    }

    $canvas lower $tmp
    $canvas lower $link

    add_to_node_link $canvas $tag1 Task_From $link
    add_to_node_link $canvas $tag2 Task_To $link
    add_to_node_link $canvas $tag1 ${link}_Task_From $tmp
    add_to_node_link $canvas $tag2 ${link}_Task_To $tmp

    add_to_link $canvas $link Task $tag1 $tag2

    bind_line $canvas $link
}

proc get_means_link_tags {tag1 tag2} {
    set t [get_link_tags $tag1 $tag2]
    lappend t meanslink$tag1$tag2 meanslink
    return $t
}

proc get_means_stub_tags {link} {
    set t [get_stub_tags $link]
    lappend t meansstub$link meansstub
    return $t
}

proc get_means_link_label_tags {link} {
    set t [get_link_label_tags $link]
    lappend t meanslinklabel$link meanslinklabel
    return $t
}

proc get_means_stub_coord {x1 y1 x2 y2 slope} {
    global Display_Link_Size PI 

#    set lx [expr $x1*0.65+$x2*0.35]
#    set lx2 [expr $x1*0.75+$x2*0.25]
#    set ly [expr $y1*0.65+$y2*0.35]
#    set ly2 [expr $y1*0.75+$y2*0.25]
### changed by mike
    set lx [expr $x1*0.50+$x2*0.50]
    set lx2 [expr $x1*0.60+$x2*0.40]
    set ly [expr $y1*0.50+$y2*0.50]
    set ly2 [expr $y1*0.60+$y2*0.40]

    set tmp $Display_Link_Size(Means)

    if {[expr $slope-0.0] != 0} {
        set dx [expr sqrt($tmp*$tmp/(1/(($slope*1.0)*$slope*1.0+0.0001)+1+0.001))]
        set dy [expr (0-1/($slope*1.0+0.001))*$dx]
    } else {
        set dx 0
        set dy $tmp
    }

    return [list [expr $lx-$dx] [expr $ly-$dy] \
            $lx2 $ly2 \
            [expr $lx+$dx] [expr $ly+$dy]]
}

proc get_means_label_coord {coord} {
    return [get_task_label_coord $coord]
}

proc create_means_ends_link {canvas tag1 tag2 {label ""}} {
    global Display_Node_Pos Display_Node_Link
    global Display_Font

    set fn $Display_Font(Link)

    set x1 [lindex $Display_Node_Pos($canvas,$tag1) 0]
    set y1 [lindex $Display_Node_Pos($canvas,$tag1) 1]

    set x2 [lindex $Display_Node_Pos($canvas,$tag2) 0]
    set y2 [lindex $Display_Node_Pos($canvas,$tag2) 1]

    set link \
      [$canvas create line $x1 $y1 \
        $x2 $y2 -width 2 \
        -fill black \
        -tag [get_means_link_tags $tag1 $tag2]]

    set coord [get_means_stub_coord $x1 $y1 $x2 $y2 \
            [compute_task_stub_slope $x1 $y1 $x2 $y2]]

    set tmp \
      [$canvas create line [lindex $coord 0] \
       [lindex $coord 1] [lindex $coord 2] [lindex $coord 3] \
       [lindex $coord 4] [lindex $coord 5] -width 2 \
       -fill black \
       -tag [get_means_stub_tags $link]]

    if {$label != ""} {
        set cod [get_means_label_coord $coord]

        set tmp2 \
                [$canvas create text [lindex $cod 0] [lindex $cod 1] \
                -text "$label" -anchor s -font $fn\
                -tag [get_means_link_label_tags $link]]

        $canvas lower $tmp2

        add_to_node_link $canvas $tag1 ${link}_Means_Label_From $tmp2
        add_to_node_link $canvas $tag2 ${link}_Means_Label_To $tmp2
    }

    $canvas lower $tmp
    $canvas lower $link

    add_to_node_link $canvas $tag1 Means_From $link
    add_to_node_link $canvas $tag2 Means_To $link
    add_to_node_link $canvas $tag1 ${link}_Means_From $tmp
    add_to_node_link $canvas $tag2 ${link}_Means_To $tmp

    add_to_link $canvas $link Means $tag1 $tag2

    bind_line $canvas $link
}

proc get_softgoal_link_tags {tag1 tag2} {
    set t [get_link_tags $tag1 $tag2] 
    lappend t softgoallink$tag1$tag2 softgoallink
    return $t
}

proc get_softgoal_stub_tags {link} {
    set t [get_stub_tags $link] 
    lappend t softgoal$link softgoalstub
    return $t
}

proc get_softgoal_link_label_tags {link} {
    set t [get_link_label_tags $link]
    lappend t softgoallinklabel$link softgoallinklabel
    return $t
}

proc get_softgoal_stub_coord {x1 y1 x2 y2 slope} {
    global Display_Link_Size PI 
####changed by mike    
### en train de travailler by mike
    set distx [expr abs($x2-$x1)]
    set disty [expr abs($y2-$y1)]
    set dist [expr $distx+$disty]

#    set width $Display_Size(Softgoal)
#    set dist2 [expr ($dist - $width)-20]
#    set alpha [ expr $dist2/$dist]
#    set beta [expr 1- $alpha]

#    set lx [expr $x1*$alpha+$x2*$beta]
#    set lx2 [expr $x1*($alpha+0.1)+$x2*($beta - 0.1)]
#    set lx3 [expr $x1*($alpha+0.05)+$x2*($beta-0.05)]
#    set ly [expr $y1*$alpha+$y2*$beta]
#    set ly2 [expr $y1*($alpha + 0.1) +$y2*($beta-0.1)]
#    set ly3 [expr $y1*($alpha + 0.05) +$y2*($beta-0.05)]

    set lx [expr $x1*0.50+$x2*0.50]
    set lx2 [expr $x1*0.60+$x2*0.40]
    set lx3 [expr $x1*0.55+$x2*0.45]
    set ly [expr $y1*0.50+$y2*0.50]
    set ly2 [expr $y1*0.60+$y2*0.40]
    set ly3 [expr $y1*0.55+$y2*0.45]

#    set lx [expr $x1*0.65+$x2*0.35]
#    set lx2 [expr $x1*0.75+$x2*0.25]
#    set lx3 [expr $x1*0.70+$x2*0.30]
#    set ly [expr $y1*0.65+$y2*0.35]
#    set ly2 [expr $y1*0.75+$y2*0.25]
#    set ly3 [expr $y1*0.70+$y2*0.30]
    set tmp $Display_Link_Size(Means)

    if {[expr $slope-0.0] != 0} {
        set dx [expr sqrt($tmp*$tmp/(1/(($slope*1.0)*$slope*1.0+0.0001)+1+0.001))]
        set dy [expr (0-1/($slope*1.0+0.001))*$dx]
    } else {
        set dx 0
        set dy $tmp
    }

    return [list [expr $lx-$dx] [expr $ly-$dy] \
            $lx2 $ly2 \
            [expr $lx+$dx] [expr $ly+$dy] \
            $lx3 $ly3]
}

proc get_softgoal_label_coord {coord} {
    return [get_task_label_coord $coord]
}

proc create_softgoal_link {canvas tag1 tag2 {label ""}} {
    global Display_Node_Pos Display_Node_Link
    global Display_Font

    set fn $Display_Font(Link)

    set x1 [lindex $Display_Node_Pos($canvas,$tag1) 0]
    set y1 [lindex $Display_Node_Pos($canvas,$tag1) 1]

    set x2 [lindex $Display_Node_Pos($canvas,$tag2) 0]
    set y2 [lindex $Display_Node_Pos($canvas,$tag2) 1]

    set link \
      [$canvas create line $x1 $y1 \
        $x2 $y2 -width 2 \
        -fill black \
        -tag [get_softgoal_link_tags $tag1 $tag2]]

    set coord [get_softgoal_stub_coord $x1 $y1 $x2 $y2 \
            [compute_task_stub_slope $x1 $y1 $x2 $y2]]

    set tmp \
      [$canvas create polygon [lindex $coord 0] \
       [lindex $coord 1] [lindex $coord 2] [lindex $coord 3] \
       [lindex $coord 4] [lindex $coord 5] [lindex $coord 6] \
       [lindex $coord 7] \
       -fill black \
       -tag [get_softgoal_stub_tags $link]]

    if {$label != ""} {
        set cod [get_softgoal_label_coord $coord]

        set tmp2 \
                [$canvas create text [lindex $cod 0] [lindex $cod 1] \
                -text "$label" -anchor s -font $fn \
                -tag [get_softgoal_link_label_tags $link]]

        $canvas lower $tmp2

        add_to_node_link $canvas $tag1 ${link}_SGoal_Label_From $tmp2
        add_to_node_link $canvas $tag2 ${link}_SGoal_Label_To $tmp2
    }

    $canvas lower $tmp
    $canvas lower $link

    add_to_node_link $canvas $tag1 SGoal_From $link
    add_to_node_link $canvas $tag2 SGoal_To $link
    add_to_node_link $canvas $tag1 ${link}_SGoal_From $tmp
    add_to_node_link $canvas $tag2 ${link}_SGoal_To $tmp

    add_to_link $canvas $link SGoal $tag1 $tag2

    bind_line $canvas $link
}

########################################################################
# Start Relationships
########################################################################

proc get_part_relation_tags {tag1 tag2} {
    set t [get_link_tags $tag1 $tag2]
    lappend t partlink$tag1$tag2 partlink
    return $t
}

proc get_part_stub_tags {link} {
    set t [get_stub_tags $link]
    lappend t part$link partstub
    return $t
}

proc get_part_relation_label_tags {link} {
    set t [get_link_label_tags $link]
    lappend t partlinklabel$link partlinklabel
    return $t
}

proc get_part_stub_coord {x1 y1 x2 y2 slope} {
    global Display_Link_Size PI

####changed by mike

    set lx [expr $x1*0.50+$x2*0.50]
    set lx2 [expr $x1*0.60+$x2*0.40
    set ly [expr $y1*0.50+$y2*0.50]
    set ly2 [expr $y1*0.60+$y2*0.40]
    
#    set lx [expr $x1*0.65+$x2*0.35]
#    set lx2 [expr $x1*0.75+$x2*0.25]
#    set ly [expr $y1*0.65+$y2*0.35]
#    set ly2 [expr $y1*0.75+$y2*0.25]

    set tmp $Display_Link_Size(Means)

    if {[expr $slope-0.0] != 0} {
        set dx [expr sqrt($tmp*$tmp/(1/(($slope*1.0)*$slope*1.0+0.0001)+1+0.001))]
        set dy [expr (0-1/($slope*1.0+0.001))*$dx]
    } else {
        set dx 0
        set dy $tmp
    }
    
    return [list [expr $lx2-$dx] [expr $ly2-$dy] \
            $lx $ly \
            [expr $lx2+$dx] [expr $ly2+$dy]]
}

proc get_part_label_coord {coord} {
    return [get_task_label_coord $coord]
}

proc create_part_relation {canvas tag1 tag2 {label ""}} {
    global Display_Node_Pos Display_Node_Link
    global Display_Font

    set fn $Display_Font(Link)

    set x1 [lindex $Display_Node_Pos($canvas,$tag1) 0]
    set y1 [lindex $Display_Node_Pos($canvas,$tag1) 1]

    set x2 [lindex $Display_Node_Pos($canvas,$tag2) 0]
    set y2 [lindex $Display_Node_Pos($canvas,$tag2) 1]

    set link \
      [$canvas create line $x1 $y1 \
        $x2 $y2 -width 2 \
        -fill black \
        -tag [get_part_relation_tags $tag1 $tag2]]

    set coord [get_part_stub_coord $x1 $y1 $x2 $y2 \
            [compute_task_stub_slope $x1 $y1 $x2 $y2]]

    set tmp \
      [$canvas create line [lindex $coord 0] \
       [lindex $coord 1] [lindex $coord 2] [lindex $coord 3] \
       [lindex $coord 4] [lindex $coord 5] -width 2 \
       -fill black \
       -tag [get_part_stub_tags $link]]

    if {$label != ""} {
        set cod [get_part_label_coord $coord]

        set tmp2 \
                [$canvas create text [lindex $cod 0] [lindex $cod 1] \
                -text "$label" -anchor s -font $fn\
                -tag [get_part_relation_label_tags $link]]

        $canvas lower $tmp2

        add_to_node_link $canvas $tag1 ${link}_Part_Label_From $tmp2
        add_to_node_link $canvas $tag2 ${link}_Part_Label_To $tmp2
    }

    $canvas lower $tmp
    $canvas lower $link

    add_to_node_link $canvas $tag1 Part_From $link
    add_to_node_link $canvas $tag2 Part_To $link
    add_to_node_link $canvas $tag1 ${link}_Part_From $tmp
    add_to_node_link $canvas $tag2 ${link}_Part_To $tmp

    add_to_link $canvas $link Part $tag1 $tag2

    bind_line $canvas $link
}


proc get_play_relation_tags {tag1 tag2} {
    set t [get_link_tags $tag1 $tag2]
    lappend t playlink$tag1$tag2 playlink
    return $t
}

proc get_play_stub_tags {link} {
    set t [get_stub_tags $link]
    lappend t play$link playstub
    return $t
}

proc get_play_relation_label_tags {link} {
    set t [get_link_label_tags $link]
    lappend t playlinklabel$link playlinklabel
    return $t
}

proc get_play_stub_coord {x1 y1 x2 y2 slope} {
    global Display_Link_Size PI

####changed by mike
    set lx [expr $x1*0.50+$x2*0.50]
    set lx2 [expr $x1*0.60+$x2*0.40
    set ly [expr $y1*0.50+$y2*0.50]
    set ly2 [expr $y1*0.60+$y2*0.40]    

#    set lx [expr $x1*0.65+$x2*0.35]
#    set lx2 [expr $x1*0.75+$x2*0.25]
#    set ly [expr $y1*0.65+$y2*0.35]
#    set ly2 [expr $y1*0.75+$y2*0.25]

    set tmp $Display_Link_Size(Means)

    if {[expr $slope-0.0] != 0} {
        set dx [expr sqrt($tmp*$tmp/(1/(($slope*1.0)*$slope*1.0+0.0001)+1+0.001))]
        set dy [expr (0-1/($slope*1.0+0.001))*$dx]
    } else {
        set dx 0
        set dy $tmp
    }
    
    return [list [expr $lx2-$dx] [expr $ly2-$dy] \
            $lx $ly \
            [expr $lx2+$dx] [expr $ly2+$dy]]
}

proc get_play_label_coord {coord} {
    return [get_task_label_coord $coord]
}

proc create_play_relation {canvas tag1 tag2 {label ""}} {
    global Display_Node_Pos Display_Node_Link
    global Display_Font

    set fn $Display_Font(Link)

    set x1 [lindex $Display_Node_Pos($canvas,$tag1) 0]
    set y1 [lindex $Display_Node_Pos($canvas,$tag1) 1]

    set x2 [lindex $Display_Node_Pos($canvas,$tag2) 0]
    set y2 [lindex $Display_Node_Pos($canvas,$tag2) 1]

    set link \
      [$canvas create line $x1 $y1 \
        $x2 $y2 -width 2 \
        -fill black \
        -tag [get_play_relation_tags $tag1 $tag2]]

    set coord [get_play_stub_coord $x1 $y1 $x2 $y2 \
            [compute_task_stub_slope $x1 $y1 $x2 $y2]]

    set tmp \
      [$canvas create line [lindex $coord 0] \
       [lindex $coord 1] [lindex $coord 2] [lindex $coord 3] \
       [lindex $coord 4] [lindex $coord 5] -width 2 \
       -fill black \
       -tag [get_play_stub_tags $link]]

    if {$label != ""} {
        set cod [get_play_label_coord $coord]

        set tmp2 \
                [$canvas create text [lindex $cod 0] [lindex $cod 1] \
                -text "$label" -anchor s -font $fn\
                -tag [get_play_relation_label_tags $link]]

        $canvas lower $tmp2

        add_to_node_link $canvas $tag1 ${link}_Play_Label_From $tmp2
        add_to_node_link $canvas $tag2 ${link}_Play_Label_To $tmp2
    }

    $canvas lower $tmp
    $canvas lower $link

    add_to_node_link $canvas $tag1 Play_From $link
    add_to_node_link $canvas $tag2 Play_To $link
    add_to_node_link $canvas $tag1 ${link}_Play_From $tmp
    add_to_node_link $canvas $tag2 ${link}_Play_To $tmp

    add_to_link $canvas $link Play $tag1 $tag2

    bind_line $canvas $link
}

proc get_cover_relation_tags {tag1 tag2} {
    set t [get_link_tags $tag1 $tag2]
    lappend t coverlink$tag1$tag2 coverlink
    return $t
}

proc get_cover_stub_tags {link} {
    set t [get_stub_tags $link]
    lappend t cover$link coverstub
    return $t
}

proc get_cover_relation_label_tags {link} {
    set t [get_link_label_tags $link]
    lappend t coverlinklabel$link coverlinklabel
    return $t
}

proc get_cover_stub_coord {x1 y1 x2 y2 slope} {
    global Display_Link_Size PI

####changed by mike
    set lx [expr $x1*0.50+$x2*0.50]
    set lx2 [expr $x1*0.60+$x2*0.40
    set ly [expr $y1*0.50+$y2*0.50]
    set ly2 [expr $y1*0.60+$y2*0.40]
    
#    set lx [expr $x1*0.65+$x2*0.35]
#    set lx2 [expr $x1*0.75+$x2*0.25]
#    set ly [expr $y1*0.65+$y2*0.35]
#    set ly2 [expr $y1*0.75+$y2*0.25]

    set tmp $Display_Link_Size(Means)

    if {[expr $slope-0.0] != 0} {
        set dx [expr sqrt($tmp*$tmp/(1/(($slope*1.0)*$slope*1.0+0.0001)+1+0.001))]
        set dy [expr (0-1/($slope*1.0+0.001))*$dx]
    } else {
        set dx 0
        set dy $tmp
    }
    
    return [list [expr $lx2-$dx] [expr $ly2-$dy] \
            $lx $ly \
            [expr $lx2+$dx] [expr $ly2+$dy]]
}

proc get_cover_label_coord {coord} {
    return [get_task_label_coord $coord]
}

proc create_cover_relation {canvas tag1 tag2 {label ""}} {
    global Display_Node_Pos Display_Node_Link
    global Display_Font

    set fn $Display_Font(Link)

    set x1 [lindex $Display_Node_Pos($canvas,$tag1) 0]
    set y1 [lindex $Display_Node_Pos($canvas,$tag1) 1]

    set x2 [lindex $Display_Node_Pos($canvas,$tag2) 0]
    set y2 [lindex $Display_Node_Pos($canvas,$tag2) 1]

    set link \
      [$canvas create line $x1 $y1 \
        $x2 $y2 -width 2 \
        -fill black \
        -tag [get_cover_relation_tags $tag1 $tag2]]

    set coord [get_cover_stub_coord $x1 $y1 $x2 $y2 \
            [compute_task_stub_slope $x1 $y1 $x2 $y2]]

    set tmp \
      [$canvas create line [lindex $coord 0] \
       [lindex $coord 1] [lindex $coord 2] [lindex $coord 3] \
       [lindex $coord 4] [lindex $coord 5] -width 2 \
       -fill black \
       -tag [get_cover_stub_tags $link]]

    if {$label != ""} {
        set cod [get_cover_label_coord $coord]

        set tmp2 \
                [$canvas create text [lindex $cod 0] [lindex $cod 1] \
                -text "$label    " -anchor s -font $fn\
                -tag [get_cover_relation_label_tags $link]]

        $canvas lower $tmp2

        add_to_node_link $canvas $tag1 ${link}_Cover_Label_From $tmp2
        add_to_node_link $canvas $tag2 ${link}_Cover_Label_To $tmp2
    }

    $canvas lower $tmp
    $canvas lower $link

    add_to_node_link $canvas $tag1 Cover_From $link
    add_to_node_link $canvas $tag2 Cover_To $link
    add_to_node_link $canvas $tag1 ${link}_Cover_From $tmp
    add_to_node_link $canvas $tag2 ${link}_Cover_To $tmp

    add_to_link $canvas $link Cover $tag1 $tag2

    bind_line $canvas $link
}

proc get_isa_relation_tags {tag1 tag2} {
    set t [get_link_tags $tag1 $tag2]
    lappend t isalink$tag1$tag2 isalink
    return $t
}

proc get_isa_stub_tags {link} {
    set t [get_stub_tags $link]
    lappend t isa$link isastub
    return $t
}

proc get_isa_relation_label_tags {link} {
    set t [get_link_label_tags $link]
    lappend t isalinklabel$link isalinklabel
    return $t
}

proc get_isa_stub_coord {x1 y1 x2 y2 slope} {
    global Display_Link_Size PI

####changed by mike
    set lx [expr $x1*0.50+$x2*0.50]
    set lx2 [expr $x1*0.60+$x2*0.40
    set ly [expr $y1*0.50+$y2*0.50]
    set ly2 [expr $y1*0.60+$y2*0.40]
    
#    set lx [expr $x1*0.65+$x2*0.35]
#    set lx2 [expr $x1*0.75+$x2*0.25]
#    set ly [expr $y1*0.65+$y2*0.35]
#    set ly2 [expr $y1*0.75+$y2*0.25]

    set tmp $Display_Link_Size(Means)

    if {[expr $slope-0.0] != 0} {
        set dx [expr sqrt($tmp*$tmp/(1/(($slope*1.0)*$slope*1.0+0.0001)+1+0.001))]
        set dy [expr (0-1/($slope*1.0+0.001))*$dx]
    } else {
        set dx 0
        set dy $tmp
    }
    
    return [list [expr $lx2-$dx] [expr $ly2-$dy] \
            $lx $ly \
            [expr $lx2+$dx] [expr $ly2+$dy]]
}

proc get_isa_label_coord {coord} {
    return [get_task_label_coord $coord]
}

proc create_isa_relation {canvas tag1 tag2 {label ""}} {
    global Display_Node_Pos Display_Node_Link
    global Display_Font

    set fn $Display_Font(Link)

    set x1 [lindex $Display_Node_Pos($canvas,$tag1) 0]
    set y1 [lindex $Display_Node_Pos($canvas,$tag1) 1]

    set x2 [lindex $Display_Node_Pos($canvas,$tag2) 0]
    set y2 [lindex $Display_Node_Pos($canvas,$tag2) 1]

    set link \
      [$canvas create line $x1 $y1 \
        $x2 $y2 -width 2 \
        -fill black \
        -tag [get_isa_relation_tags $tag1 $tag2]]

    set coord [get_isa_stub_coord $x1 $y1 $x2 $y2 \
            [compute_task_stub_slope $x1 $y1 $x2 $y2]]

    set tmp \
      [$canvas create line [lindex $coord 0] \
       [lindex $coord 1] [lindex $coord 2] [lindex $coord 3] \
       [lindex $coord 4] [lindex $coord 5] -width 2 \
       -fill black \
       -tag [get_isa_stub_tags $link]]

    if {$label != ""} {
        set cod [get_isa_label_coord $coord]

        set tmp2 \
                [$canvas create text [lindex $cod 0] [lindex $cod 1] \
                -text "$label" -anchor s -font $fn\
                -tag [get_isa_relation_label_tags $link]]

        $canvas lower $tmp2

        add_to_node_link $canvas $tag1 ${link}_Isa_Label_From $tmp2
        add_to_node_link $canvas $tag2 ${link}_Isa_Label_To $tmp2
    }

    $canvas lower $tmp
    $canvas lower $link

    add_to_node_link $canvas $tag1 Isa_From $link
    add_to_node_link $canvas $tag2 Isa_To $link
    add_to_node_link $canvas $tag1 ${link}_Isa_From $tmp
    add_to_node_link $canvas $tag2 ${link}_Isa_To $tmp

    add_to_link $canvas $link Isa $tag1 $tag2

    bind_line $canvas $link
}

proc get_occupy_relation_tags {tag1 tag2} {
    set t [get_link_tags $tag1 $tag2]
    lappend t occupylink$tag1$tag2 occupylink
    return $t
}

proc get_occupy_stub_tags {link} {
    set t [get_stub_tags $link]
    lappend t occupy$link occupystub
    return $t
}

proc get_occupy_relation_label_tags {link} {
    set t [get_link_label_tags $link]
    lappend t occupylinklabel$link occupylinklabel
    return $t
}

proc get_occupy_stub_coord {x1 y1 x2 y2 slope} {
    global Display_Link_Size PI

####changed by mike
    set lx [expr $x1*0.50+$x2*0.50]
    set lx2 [expr $x1*0.60+$x2*0.40
    set ly [expr $y1*0.50+$y2*0.50]
    set ly2 [expr $y1*0.60+$y2*0.40]

#    set lx [expr $x1*0.65+$x2*0.35]
#    set lx2 [expr $x1*0.75+$x2*0.25]
#    set ly [expr $y1*0.65+$y2*0.35]
#    set ly2 [expr $y1*0.75+$y2*0.25]

    set tmp $Display_Link_Size(Means)

    if {[expr $slope-0.0] != 0} {
        set dx [expr sqrt($tmp*$tmp/(1/(($slope*1.0)*$slope*1.0+0.0001)+1+0.001))]
        set dy [expr (0-1/($slope*1.0+0.001))*$dx]
    } else {
        set dx 0
        set dy $tmp
    }
    
    return [list [expr $lx2-$dx] [expr $ly2-$dy] \
            $lx $ly \
            [expr $lx2+$dx] [expr $ly2+$dy]]
}

proc get_occupy_label_coord {coord} {
    return [get_task_label_coord $coord]
}

proc create_occupy_relation {canvas tag1 tag2 {label ""}} {
    global Display_Node_Pos Display_Node_Link
    global Display_Font

    set fn $Display_Font(Link)

    set x1 [lindex $Display_Node_Pos($canvas,$tag1) 0]
    set y1 [lindex $Display_Node_Pos($canvas,$tag1) 1]

    set x2 [lindex $Display_Node_Pos($canvas,$tag2) 0]
    set y2 [lindex $Display_Node_Pos($canvas,$tag2) 1]

    set link \
      [$canvas create line $x1 $y1 \
        $x2 $y2 -width 2 \
        -fill black \
        -tag [get_occupy_relation_tags $tag1 $tag2]]

    set coord [get_occupy_stub_coord $x1 $y1 $x2 $y2 \
            [compute_task_stub_slope $x1 $y1 $x2 $y2]]

    set tmp \
      [$canvas create line [lindex $coord 0] \
       [lindex $coord 1] [lindex $coord 2] [lindex $coord 3] \
       [lindex $coord 4] [lindex $coord 5] -width 2 \
       -fill black \
       -tag [get_occupy_stub_tags $link]]

    if {$label != ""} {
        set cod [get_occupy_label_coord $coord]

        set tmp2 \
                [$canvas create text [lindex $cod 0] [lindex $cod 1] \
                -text "$label" -anchor s -font $fn\
                -tag [get_occupy_relation_label_tags $link]]

        $canvas lower $tmp2

        add_to_node_link $canvas $tag1 ${link}_Occupy_Label_From $tmp2
        add_to_node_link $canvas $tag2 ${link}_Occupy_Label_To $tmp2
    }

    $canvas lower $tmp
    $canvas lower $link

    add_to_node_link $canvas $tag1 Occupy_From $link
    add_to_node_link $canvas $tag2 Occupy_To $link
    add_to_node_link $canvas $tag1 ${link}_Occupy_From $tmp
    add_to_node_link $canvas $tag2 ${link}_Occupy_To $tmp

    add_to_link $canvas $link Occupy $tag1 $tag2

    bind_line $canvas $link
}
