# telos: parse the latest telos syntax
#
#
#   "Last Modified: Fri Dec  9 18:46:58 1994 by Bryan M. Kramer"
#

set commentpat {\(\*.*\*\)}
set commentline {^[	 ]*\(\*.*\*\)}

proc getline {stream} {
    global eof line_no first_root
    if {[gets $stream line] >= 0} {
        incr line_no
        set result $line
    } else {
        set eof 1
        set result ""
    }
}

proc attr_inter {list1 list2} {
    set result 0
    catch {
        foreach x $list1 {
            if {[lsearch $list2 $x] >= 0} {
                set result 1
                return 1
            }
        }
    }
    return $result
}
 
proc get_attr_val {obj cat} {
    global slots
    set result {}
    set slots {}

    foreach slot [kb attributes $obj] {
        if {[attr_inter [kb instance_names $slot] $cat]} {
	    lappend slots $slot
            lappend result [lindex [kb value $slot] 0]
        }
    }
    return [list $result $slots]
}

proc kb_create {type name x y {wrt ""}} {
    kb create $name Token
    kb instance_of $name "${type}Class"
    if {$wrt != ""} {
	kb attr2 $name wrt {} $wrt
    }
}

proc kb_coord {canvas name x y} {
    global KB_Slot_ID

    if {[info exists KB_Slot_ID($canvas,$name,x)]} {
	kb value $KB_Slot_ID($canvas,$name,x) $x
	kb value $KB_Slot_ID($canvas,$name,y) $y
    } else {
	kb attr2 $name coords "x" $x
	kb attr2 $name coords "y" $y
    }
}
    
proc kb_add {type name val} {
    kb attr2 $name depend_to $type $val
}

proc add_object {canvas type obj} {
    global Exist_Object

    if {[info exists Exist_Object($canvas,$type)]} {
	set t $Exist_Object($canvas,$type)
	lappend t $obj
	set Exist_Object($canvas,$type) $t
    } else {
	set Exist_Object($canvas,$type) "$obj"
    }
}
  
proc check_design {canvas obj {classes ""}} {
    foreach class $classes {
	if {$class == "ActorClass"} {
	    add_object $canvas Actor $obj
	} elseif {$class == "AgentClass"} {
	    add_object $canvas Agent $obj	
 	} elseif {$class == "RoleClass"} {
	    add_object $canvas Role $obj
	} elseif {$class == "PositionClass"} {
	    add_object $canvas Position $obj
	} elseif {$class == "ResourceClass"} {
	    add_object $canvas Resource $obj
	} elseif {$class == "TaskClass"} {
	    add_object $canvas Task $obj
	} elseif {$class == "GoalClass"} {
	    add_object $canvas Goal $obj
	} elseif {$class == "SoftgoalClass"} {
	    add_object $canvas Softgoal $obj
	} elseif {$class == "LabelClass"} {
	    add_object $canvas Label $obj
	} elseif {$class == "AlternativeClass"} {
            add_object $canvas Alternative $obj
	}
    }
}
	    
proc built_in {name} {
    if {$name == "SRElementClass" || \
	$name == "ActorClass" || \
	$name == "AgentClass" || \
	$name == "RoleClass" || \
	$name == "PositionClass" || \
	$name == "TaskClass" || \
	$name == "ResourceClass" || \
	$name == "GoalClass" || \
	$name == "SoftgoalClass" || \
	$name == "LabelClass" || \
	$name == "AlternativeClass"} {
	return 1
    } else {
	return 0
    }
}

proc parse_design_file {canvas file} {
    global line_no eof
    global commentpat commentline

    set stream 0
    set eof 0

    catch {set stream [open "$file" r]} msg
    if { $stream == 0 } {
	puts stderr "Error: $msg"
	exit
    }

    set line_no 0
    set line [getline $stream]
    set obj ""
    set cat ""
    set inwith 0
    set instring 0
    set repeat 0

    while {$eof == 0} {
	if {[regexp "$commentline" $line]} {
	} elseif {[regexp {^[	 ]*MetaMetaClass[	 ]*([^	 ]*)([	 ()].*)?$} $line all name]} {
	    set name [string trim $name |]
	    if {[built_in $name] && [kb exists $name]} {
		set repeat 1
	    } else {
		puts "metametaclass $name"
		set obj $name
		kb create $obj MetaMetaClass
	    }
	} elseif {[regexp {^[	 ]*MetaClass[	 ]*([^	 ]*)([	 ()].*)?$} $line all name]} {
	    set name [string trim $name |]
	    if {[built_in $name] && [kb exists $name]} {
		set repeat 1
	    } else {
		puts "metaclass $name"
		set obj $name
		kb create $obj MetaClass
	    }
	} elseif {[regexp {^[	 ]*SimpleClass[	 ]*([^	 ]*)([	 ()].*)?$} $line all name]} {
	    set name [string trim $name |]
	    if {[built_in $name] && [kb exists $name]} {
		set repeat 1
	    } else {
		puts "simpleclass $name"
		set obj $name
		kb create $obj SimpleClass
	    }
	} elseif {[regexp {^[	 ]*Token[	 ]*([^	 ]*)([	 ()].*)?$} $line all name]} {
	    set name [string trim $name |]
	    if {[built_in $name] && [kb exists $name]} {
		set repeat 1
	    } else {
		puts "token $name"
		set obj $name
		kb create $obj Token
	    }
	} elseif {[regexp {^[	 ]*Class[	 ]*([^	 ]*)([	 ()].*)?$} $line all name]} {
	    # ignore these
	    set repeat 1
	} elseif {[regexp {^[	 ]*end[()	 ]*} $line]} {
	    # object is complete
	    if {! $repeat} {
		puts "object $obj is complete"
	    }
	    puts ""
	    set obj ""
	    set inwith 0
	    set cat ""
	    set instring 0
	    set repeat 0
	} elseif {[regexp {^[	 ]*in[	 ]+([^()]*)} $line all classes]} {
	    if {! $repeat} {
		regsub , $classes {} classes
		set classes [split $classes { }]
		foreach x $classes {
		    check_design $canvas $obj $x
		    puts "instance of $x"
		    kb instance_of $obj $x
		}
	    }
	} elseif {[regexp {^[	 ]*is[aA][	 ]+([^()]*)} $line all classes]} {
	    if {! $repeat} {
		puts "isa $classes"
		kb isa $obj $classes
	    }
	} elseif {! $inwith && [regexp {^[	 ]*with[()	 ]*} $line]} {
	    if {! $repeat} {
		puts "with"
		set inwith 1
	    }
	} elseif {! $inwith} {
	    # cant do anything if not in with
	} elseif {[regexp {^[	 ]*([^	 :;()][^	 ():;]*)$} $line all catnames]} {
	    if {! $repeat} {
		puts "category $catnames"
		set cat $catnames
	    }
	} elseif {[regexp {^[	 ]*([^	 :;()][^	 ():;]*)[	 ]*\(\*$} $line all catnames]} {
	    if {! $repeat} {
		puts "category $catnames"
		set cat $catnames
	    }
	} elseif {$cat == ""} {
	} elseif {[regexp {^[	 ]*([^	 :;()][^	 ():;]*)?[	 ]*:[	 ]*([^:;""][^:;""]*)[	 ]*;[	 ]*$} $line all label value]} {
	    if {! $repeat} {
		set label [string trim $label |]
		set value [string trim $value |]
		puts "label $label with value $value"
		kb attr2 $obj $cat $label $value
	    }
	} elseif {[regexp {^[	 ]*([^	 :;()][^	 ():;]*)?[	 ]*:[	 ]*([^:;""][^:;""]*)[	 ]*$} $line all label value]} {
	    if {! $repeat} {
		set label [string trim $label |]
		set value [string trim $value |]
		puts "label $label with value $value"
		kb attr2 $obj $cat $label $value
	    }
	}
	set line [getline $stream]
    }

    if {$file != "class.tel"} {
	complete_file_design $canvas $file
    }
}
